// Copyright (c) 2017-2023 Advanced Micro Devices, Inc. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef HIPRAND_ROCM_H_
#define HIPRAND_ROCM_H_

#include <rocrand/rocrand.h>

typedef struct rocrand_generator_base_type hiprandGenerator_st;

typedef struct rocrand_discrete_distribution_st hiprandDiscreteDistribution_st;

/// \addtogroup hipranddevice
/// @{

/// \brief Convenience typedef for Sobol 32 direction vector array
typedef unsigned int       hiprandDirectionVectors32_t[32];

/// \brief Convenience typedef for Sobol 64 direction vector array
typedef unsigned long long hiprandDirectionVectors64_t[64];

/// @} // end of group hipranddevice

#endif // HIPRAND_ROCM_H_
